"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.nexusPrisma = void 0;
const nexus_1 = require("nexus");
const builder_1 = require("./builder");
const colors_1 = require("./colors");
/**
 * Create a nexus-plugin-prisma instance to be passed into the Nexus plugins array.
 *
 * The nexus plugin extends the Nexus DSL with `t.model` and `t.crud`. These
 * allow you to expose data from your Prisma layer and operations against that
 * data. Once this plugin is installed, you can do things such as:
 *
 * ```ts
 * objectType({
 *   name: 'User',
 *   definition(t) {
 *     t.model.id()
 *     t.model.email()
 *   }
 * })
 *
 * queryType({
 *   definition (t) {
 *      t.crud.user()
 *      t.crud.users({ filtering: true, ordering: true })
 *    }
 * })
 * ```
 *
 * You must ensure the Prisma Client JS has been generated prior as it provides a
 * data representation of the available models and CRUD operations against them.
 *
 * Like Nexus, NexusPrisma has its own Typegen. By default It will be generated
 * synchronously when Nexus runs the plugin, _if_ `process.env.NODE_ENV` is
 * `undefined` or `"development"`. Typegen can be explicitly enabled or disabled
 * via the shouldGenerateArtifacts option. This mirrors Nexus' own typegen
 * approach. This system will change once the Nexus plugin system has first
 * class support for typegen.
 */
function nexusPrisma(options) {
    const allOptions = options !== null && options !== void 0 ? options : {};
    let wasCrudUsedButDisabled = null;
    return nexus_1.plugin({
        name: 'nexus-plugin-prisma',
        onInstall: (nexusBuilder) => {
            const { types, wasCrudUsedButDisabled: wasCrudUsed } = builder_1.build(Object.assign(Object.assign({}, allOptions), { nexusBuilder }));
            wasCrudUsedButDisabled = wasCrudUsed;
            types.forEach(nexusBuilder.addType);
        },
        onBeforeBuild() {
            if ((wasCrudUsedButDisabled === null || wasCrudUsedButDisabled === void 0 ? void 0 : wasCrudUsedButDisabled()) === true) {
                console.log(`\
${colors_1.colors.yellow('Warning')}: ${colors_1.colors.green('t.crud')} ${colors_1.colors.yellow('is an experimental feature with many practical limitations. You must explicitly enable it before using.')}
Please add ${colors_1.colors.green(`experimentalCRUD: true`)} in the ${colors_1.colors.green('nexusSchemaPrisma()')} constructor if you still wish to enable it.`);
            }
        },
    });
}
exports.nexusPrisma = nexusPrisma;
//# sourceMappingURL=plugin.js.map